# TODO: use a more sophisticated system for this.
# When do quntities acquere units?
# - in config file?
# - when config is loaded?
# - when additional parameters are computed?

"""

The basic units are :
    centimeter              (centimeter)
    second                  (second)
    Mega electron Volt      (MeV)
    positron charge         (e)
    Kelvin                  (kelvin)
    the amount of substance (mole)
    luminous intensity      (candela)
    radian                  (radian)
    steradian               (steradian)

"""

# distance
km = 1.e5
m = 1.e2
cm = 1.
mm = 1.e-1
um = 1.e-4
nm = 1.e-7

# time
h = 3600.
m = 60.
s = 1.
ms = 1.e-3
us = 1.e-6
ns = 1.e-9

# energy
GeV = 1.e3
MeV = 1.
keV = 1.e-3
eV = 1.e-6
meV = 1.e-9



import gampixpy

import os
import yaml
import numpy as np

class Config (dict):
    def __init__(self, config_filename):
        self.config_filename = config_filename

        self.parse_config()
        self.compute_derived_parameters()
        
    def parse_config(self):
        # parse a config (yaml) file and store values
        # internally as a dict
        with open(self.config_filename) as config_file:
            for key, value in yaml.load(config_file, Loader = yaml.FullLoader).items():
                self[key] = value
        
        return

class DetectorConfig (Config):
    def compute_derived_parameters(self):
        # compute any required parameters from
        # those specified in the YAML

        return

class PhysicsConfig (Config):
    def compute_derived_parameters(self):
        # compute drift velocity, diffusion model

        # Using the mobility model described in https://arxiv.org/abs/1508.07059
        a0 = self['charge_drift']['mobility_parameter_a0'] # cm^2/V/s
        a1 = self['charge_drift']['mobility_parameter_a1']
        a2 = self['charge_drift']['mobility_parameter_a2']
        a3 = self['charge_drift']['mobility_parameter_a3']
        a4 = self['charge_drift']['mobility_parameter_a4']
        a5 = self['charge_drift']['mobility_parameter_a5']
        T0 = self['charge_drift']['mobility_reference_T0']

        b0 = self['charge_drift']['mobility_parameter_b0'] # eV
        b1 = self['charge_drift']['mobility_parameter_b1']
        b2 = self['charge_drift']['mobility_parameter_b2']
        b3 = self['charge_drift']['mobility_parameter_b3']
        T1 = self['charge_drift']['mobility_reference_T1']

        E = self['charge_drift']['drift_field'] # kV/cm
        T = self['material']['temperature'] # K
        
        numerator = a0 + a1*E + a2*np.power(E, 3./2) + a3*np.power(E, 5./2)
        denominator = 1 + (a1/a0)*E + a4*np.power(E, 2) + a5*np.power(E, 3)
        mu = numerator*np.power(T/T0, -3./2)/denominator # cm^2/V/s
        # mu *= 10*10 # mm^2/V/s
        # mu /= 1.e9 # mm^2/V/ns
        
        dMudE = ((a1 + 3./2*a2*np.power(E, 1./2) + 5./2*a3*np.power(E, 3./2))*denominator - numerator*(a1/a0 + 2*a4*E + 3*a5*np.power(E, 2)))*np.power(T/T0, -3./2)/np.power(denominator, 2) # cm^3/V/kV/s
        dMudE /= 1000.
        # dMudE *= 10*10*10 # mm^3/V^2/s
        # dMudE /= 1.e9 # mm^3/V^2/ns

        e = (b0 + b1*E + b2*np.power(E, 2))*(T/T1)/(1 + (b1/b0)*E + b3*np.power(E, 2)) # eV
        e /= 1.e6 # MeV

        R = 1 + E/mu*dMudE*1000 # diffusion ratio
        
        self['charge_drift']['electron_mobility'] = mu # cm^2/V/s
        self['charge_drift']['drift_energy'] = e # MeV
        self['charge_drift']['diffusion_longitudinal'] = mu*e*1.e6 # cm^2/s
        self['charge_drift']['diffusion_ratio'] = R
        self['charge_drift']['diffusion_transverse'] = mu*e/R*1.e6 # cm^2/s
        self['charge_drift']['drift_speed'] = mu*E*1.e3 # cm/s

        return

class ReadoutConfig (Config):
    def compute_derived_parameters(self):
        # compute any required parameters from
        # those specified in the YAML

        # need to handle the case where the pitch doesn't evenly divide the span of the anode
        self['n_pixels_x'] = int((self['anode']['x_range'][1] - self['anode']['x_range'][0])/self['pixels']['pitch'])
        self['n_pixels_y'] = int((self['anode']['y_range'][1] - self['anode']['y_range'][0])/self['pixels']['pitch'])

        self['n_tiles_x'] = int((self['anode']['x_range'][1] - self['anode']['x_range'][0])/self['coarse_tiles']['pitch'])
        self['n_tiles_y'] = int((self['anode']['y_range'][1] - self['anode']['y_range'][0])/self['coarse_tiles']['pitch'])

        self['tile_volume_edges'] = (np.linspace(self['anode']['x_range'][0],
                                                 self['anode']['x_range'][1],
                                                 self['n_tiles_x']+1),
                                     np.linspace(self['anode']['y_range'][0],
                                                 self['anode']['y_range'][1],
                                                 self['n_tiles_y']+1))
        return

default_detector_params = DetectorConfig(os.path.join(gampixpy.__path__[0],
                                                      'detector_config',
                                                      'default.yaml'))
default_physics_params = PhysicsConfig(os.path.join(gampixpy.__path__[0],
                                                    'physics_config',
                                                    'default.yaml'))
default_readout_params = ReadoutConfig(os.path.join(gampixpy.__path__[0],
                                                    'readout_config',
                                                    'default.yaml'))



from gampixpy.config import default_detector_params, default_physics_params, default_readout_params
from gampixpy.readout_objects import PixelSample, CoarseGridSample

import numpy as np
import torch

class ReadoutModel:
    def __init__(self, readout_config = default_readout_params):
        self.readout_config = readout_config

    def electronics_simulation(self, track, verbose = True, **kwargs):
        # apply the electronics response simulation to
        # a track containing a drifted charge sample

        # time of first charge arrival
        # this is the start point for coarse current building
        try:
            self.clock_start_time = torch.min(track.drifted_track['times'])
        except RuntimeError:
            print ("Encountered empty event!")
            message = """This is most often because the event lies totally outside of the specified detector volume"""
            print (message)
            print ("skipping...")

        if verbose:
            print ("simulating coarse grid...")
        # do coarse grid binning/time series formation
        coarse_tile_timeseries = self.tile_current_builder(track, **kwargs)
        if verbose:
            print ("coarse time series built")
            print ("coarse hit finding...")
        # find hits on coarse grid
        self.coarse_tile_hits = self.tile_hit_finding(track, coarse_tile_timeseries, **kwargs)

        if verbose:
            print ("simulating fine grid...")
        fine_pixel_timeseries = self.pixel_current_builder(track, self.coarse_tile_hits, **kwargs)
        if verbose:
            print ("pixel time series built")
            print ("pixel hit finding...")
        # find hits on fine pixels
        self.fine_pixel_hits = self.pixel_hit_finding(track, fine_pixel_timeseries, **kwargs)

        return 

    def point_sample_tile_current(self, tile_coord, track, sample_mask):
        """
        return a sparse timeseries for each input position
        positions are relative to the tile center
        """
        # for now, return the entire charge with the time of arrival
        # really, should return a sparse timeseries informed by FEM
        # probably loaded from a table and interpolated
        # shape will be (2, N_point_samples, N_time_samples)
        # the first axis is (timestamps, charge mass)
        # the timestamps represent the time (relative to event t0)
        # at the beginning of a small interval (ideally, much smaller than the clock)
        # and charge is the integrated current across that interval

        # sum(result[1,:,:], dim = -1) == track.drifted_track['charge']
        

        position = track.drifted_track['position'][sample_mask]
        charge = track.drifted_track['charge'][sample_mask]
        time = track.drifted_track['times'][sample_mask]
        pitch = self.readout_config['coarse_tiles']['pitch']
        
        lands_on_tile_of_interest = position[:,0] - tile_coord[0] > -0.5*pitch
        lands_on_tile_of_interest *= position[:,0] - tile_coord[0] < 0.5*pitch
        lands_on_tile_of_interest *= position[:,1] - tile_coord[1] > -0.5*pitch
        lands_on_tile_of_interest *= position[:,1] - tile_coord[1] < 0.5*pitch

        induced_charge = torch.where(lands_on_tile_of_interest,
                                     charge,
                                     torch.zeros(1),
                                     )
        return torch.stack((time, induced_charge))[:,:,None] 
    
    def compose_tile_currents(self, sparse_current_series):
        """
        combine multiple sparse timeseries into a single dense timeseries
        """
        last_charge_arrival_time = torch.max(sparse_current_series[0,:,:])
        n_clock_ticks = torch.ceil((last_charge_arrival_time - self.clock_start_time)/self.readout_config['coarse_tiles']['clock_interval']).int()

        arrival_time_bin_edges = torch.linspace(self.clock_start_time,
                                                self.clock_start_time + n_clock_ticks*self.readout_config['coarse_tiles']['clock_interval'],
                                                n_clock_ticks + 1,
                                                )

        # find the induced charge which falls into each clock bin
        induced_charge, _ = torch.histogram(sparse_current_series[0,:,:],
                                            weight = sparse_current_series[1,:,:],
                                            bins = arrival_time_bin_edges)

        return arrival_time_bin_edges[:-1], induced_charge
        
    def tile_receptive_field(self, tile_coord, track, n_neighbor_tiles = 0, **kwargs):
        """
        For each tile, select the track samples within the
        tile's receptive field
        """

        position = track.drifted_track['position']
        pitch = self.readout_config['coarse_tiles']['pitch']

        sample_mask = position[:,0] - tile_coord[0] > -(n_neighbor_tiles + 0.5)*pitch
        sample_mask *= position[:,0] - tile_coord[0] < (n_neighbor_tiles + 0.5)*pitch
        sample_mask *= position[:,1] - tile_coord[1] > -(n_neighbor_tiles + 0.5)*pitch
        sample_mask *= position[:,1] - tile_coord[1] < (n_neighbor_tiles + 0.5)*pitch

        return sample_mask 

    def tile_current_builder(self, track, **kwargs):
        """
        Build the current timeseries for each tile
        """
        min_tile = torch.tensor([self.readout_config['anode']['x_range'][0],
                                 self.readout_config['anode']['y_range'][0],
                                 ])
        spacing = self.readout_config['coarse_tiles']['pitch']
        tile_ind = torch.div(track.drifted_track['position'][:,:2] - min_tile, spacing).int()

        inside_anode_mask = (torch.min(tile_ind, axis = -1)[0] >= 0) 
        inside_anode_mask *= tile_ind[:, 0] < self.readout_config['n_tiles_x'] 
        inside_anode_mask *= tile_ind[:, 1] < self.readout_config['n_tiles_y']

        tile_ind = tile_ind[inside_anode_mask]

        z_series = track.drifted_track['position'][inside_anode_mask,2]
        t_series = track.drifted_track['times'][inside_anode_mask]
        charge_series = track.drifted_track['charge'][inside_anode_mask]

        coarse_grid_timeseries = {}
        unique_tile_indices = torch.unique(tile_ind, dim = 0)

        for this_tile_ind in unique_tile_indices:
            tile_center = torch.tensor([self.readout_config['tile_volume_edges'][i][this_tile_ind[i]] + 0.5*self.readout_config['coarse_tiles']['pitch']
                                        for i in range(2)]).T
            tile_coord = (round(float(tile_center[0]), 3),
                          round(float(tile_center[1]), 3))

            sample_mask = self.tile_receptive_field(tile_coord, track)
            tile_sample_current_series = self.point_sample_tile_current(tile_coord,
                                                                        track,
                                                                        sample_mask)
            tile_current_series = self.compose_tile_currents(tile_sample_current_series)
            coarse_grid_timeseries[tile_coord] = tile_current_series

        return coarse_grid_timeseries

    def point_sample_pixel_current(self, pixel_coord, track, sample_mask):
        """
        return a sparse timeseries for each input position
        positions are relative to the pixel center
        """
        # for now, return the entire charge with the time of arrival
        # really, should return a sparse timeseries informed by FEM
        # probably loaded from a table and interpolated
        # shape will be (2, N_point_samples, N_time_samples)
        # the first axis is (timestamps, charge mass)
        # the timestamps represent the time (relative to event t0)
        # at the beginning of a small interval (ideally, much smaller than the clock)
        # and charge is the integrated current across that interval

        # sum(result[1,:,:], dim = -1) == track.drifted_track['charge']
        
        position = track.drifted_track['position'][sample_mask]
        charge = track.drifted_track['charge'][sample_mask]
        time = track.drifted_track['times'][sample_mask]
        pitch = self.readout_config['pixels']['pitch']

        lands_on_pixel_of_interest = position[:,0] - pixel_coord[0] > -0.5*pitch
        lands_on_pixel_of_interest *= position[:,0] - pixel_coord[0] < 0.5*pitch
        lands_on_pixel_of_interest *= position[:,1] - pixel_coord[1] > -0.5*pitch
        lands_on_pixel_of_interest *= position[:,1] - pixel_coord[1] < 0.5*pitch

        induced_charge = torch.where(lands_on_pixel_of_interest,
                                     charge,
                                     torch.zeros(1),
                                     )
        return torch.stack((time, induced_charge))[:,:,None] 

    def compose_pixel_currents(self, sparse_current_series, coarse_cell_hit):
        """
        combine multiple sparse timeseries into a single dense timeseries
        """
        cell_clock_start = coarse_cell_hit.coarse_measurement_time
        cell_clock_end = coarse_cell_hit.coarse_measurement_time + self.readout_config['coarse_tiles']['clock_interval']*self.readout_config['coarse_tiles']['integration_length']
        
        n_clock_ticks = int((cell_clock_end - cell_clock_start)/self.readout_config['pixels']['clock_interval'])
        arrival_time_bin_edges = torch.linspace(cell_clock_start,
                                                cell_clock_end,
                                                n_clock_ticks + 1,
                                                )

        # find the induced charge which falls into each clock bin
        induced_charge, _ = torch.histogram(sparse_current_series[0,:,:],
                                            weight = sparse_current_series[1,:,:],
                                            bins = arrival_time_bin_edges)

        return arrival_time_bin_edges[:-1], induced_charge

    def pixel_receptive_field(self, pixel_coord, track, n_neighbor_pixels = 0, **kwargs):
        """
        For each tile, select the track samples within the
        tile's receptive field
        """

        position = track.drifted_track['position']
        pitch = self.readout_config['pixels']['pitch']

        sample_mask = position[:,0] - pixel_coord[0] > -(n_neighbor_pixels + 0.5)*pitch
        sample_mask *= position[:,0] - pixel_coord[0] < (n_neighbor_pixels + 0.5)*pitch
        sample_mask *= position[:,1] - pixel_coord[1] > -(n_neighbor_pixels + 0.5)*pitch
        sample_mask *= position[:,1] - pixel_coord[1] < (n_neighbor_pixels + 0.5)*pitch

        return sample_mask 
    
    def pixel_current_builder(self, track, coarse_tile_hits, **kwargs):
        pixel_timeseries = {}
        
        tile_pitch = self.readout_config['coarse_tiles']['pitch']
        pixel_pitch = self.readout_config['pixels']['pitch']
        
        for this_coarse_hit in coarse_tile_hits:

            cell_center_xy = this_coarse_hit.coarse_cell_pos # may change
            cell_trigger_t = this_coarse_hit.coarse_measurement_time
            
            x_bounds = [cell_center_xy[0] - 0.5*tile_pitch,
                        cell_center_xy[0] + 0.5*tile_pitch]
            y_bounds = [cell_center_xy[1] - 0.5*tile_pitch,
                        cell_center_xy[1] + 0.5*tile_pitch]
            t_bounds = [cell_trigger_t,
                        cell_trigger_t + self.readout_config['coarse_tiles']['clock_interval']*self.readout_config['coarse_tiles']['integration_length']]

            in_cell_mask = track.drifted_track['position'][:,0] >= x_bounds[0]
            in_cell_mask *= track.drifted_track['position'][:,0] < x_bounds[1]
            in_cell_mask *= track.drifted_track['position'][:,1] >= y_bounds[0]
            in_cell_mask *= track.drifted_track['position'][:,1] < y_bounds[1]
            in_cell_mask *= track.drifted_track['times'] >= t_bounds[0]
            in_cell_mask *= track.drifted_track['times'] < t_bounds[1]

            in_cell_positions = track.drifted_track['position'][in_cell_mask]
            in_cell_charges = track.drifted_track['charge'][in_cell_mask]
            
            min_pixel = torch.tensor([x_bounds[0],
                                      y_bounds[0],
                                      ])
            n_pixels_x = int((x_bounds[1] - x_bounds[0])/pixel_pitch)
            n_pixels_y = int((y_bounds[1] - y_bounds[0])/pixel_pitch)
            pixel_volume_edges = (torch.linspace(x_bounds[0], x_bounds[1], n_pixels_x+1),
                                  torch.linspace(y_bounds[0], y_bounds[1], n_pixels_y+1))
            
            z_series = in_cell_positions[:,2]
            t_series = track.drifted_track['times'][in_cell_mask]
            charge_series = in_cell_charges

            # generate a unique id for each pixel within this coarse hit
            # so that hits from 
            pixel_ind = torch.div(in_cell_positions[:,[0, 1]] - min_pixel, pixel_pitch).int()
            # pixel_hash = np.array([hash(tuple(ind)+tuple(cell_center_xy)) for ind in pixel_ind])

            # unique_pixel_hashes, unique_pixel_key = np.unique(pixel_hash, return_index = True)
            # unique_pixel_indices = pixel_ind[unique_pixel_key]
            unique_pixel_indices = torch.unique(pixel_ind, dim = 0)
            
            # for this_pixel_hash, this_pixel_ind in zip(unique_pixel_hashes, unique_pixel_indices):
            for this_pixel_ind in unique_pixel_indices:
                pixel_center = torch.tensor([pixel_volume_edges[i][this_pixel_ind[i]] + 0.5*pixel_pitch
                                             for i in range(2)]).T
                pixel_coord = (round(float(pixel_center[0]), 3),
                               round(float(pixel_center[1]), 3))

                # sample_mask = torch.all(pixel_ind == this_pixel_ind, dim = 1)
                sample_mask = self.pixel_receptive_field(pixel_coord, track)
                pixel_sample_current_series = self.point_sample_pixel_current(pixel_coord,
                                                                              track,
                                                                              sample_mask)
                pixel_current_series = self.compose_pixel_currents(pixel_sample_current_series, this_coarse_hit)
                pixel_timeseries[(pixel_coord[0],
                                  pixel_coord[1],
                                  cell_trigger_t)] = pixel_current_series
                                           
        return pixel_timeseries

class GAMPixModel (ReadoutModel):
    def tile_hit_finding(self, track, tile_timeseries, nonoise = False, **kwargs):
        """
        This method of hit finding simply looks for a quantity
        of charge above threshold within a given z-bin
        (corresponding to a clock_period*integration_length)
        """
        # TODO: fix logic for integration_length > 1
        hits = []
        
        for tile_center, timeseries in tile_timeseries.items():
            time_ticks, interval_charge = timeseries
            
            hold_length = self.readout_config['coarse_tiles']['integration_length']            
            
            # search along the bins until no more threshold crossings
            no_more_hits = False
            while not no_more_hits:
                # window_charge = np.convolve(interval_charge,
                #                             np.ones(hold_length))
                window_charge = torch.conv_tbc(interval_charge[:,None,None],
                                               torch.ones(hold_length,1,1),
                                               bias = torch.zeros(1),
                                               pad = hold_length-1)[:,0,0]
                window_charge = window_charge[hold_length-1:]
                
                threshold = self.readout_config['coarse_tiles']['noise']*self.readout_config['coarse_tiles']['threshold_sigma']

                threshold_crossing_mask = window_charge > threshold
                threshold_crossing_mask *= interval_charge > 0

                if torch.any(threshold_crossing_mask):
                    hit_index = threshold_crossing_mask.nonzero()[0][0]
                    
                    threshold_crossing_t = time_ticks[hit_index]
                    threshold_crossing_z = threshold_crossing_t*1.6e5 # is there a better way to do this?
                    # TODO: also, get that from physics params
                    
                    threshold_crossing_charge = window_charge[hit_index]
                    if not nonoise:
                        threshold_crossing_charge += torch.poisson(torch.tensor(self.readout_config['coarse_tiles']['noise']).float())

                    interval_charge[:hit_index+hold_length] = 0

                    hits.append(CoarseGridSample(tile_center,
                                                 threshold_crossing_t.item(),
                                                 threshold_crossing_z.item(),
                                                 threshold_crossing_charge.item()))
                else:
                    no_more_hits = True

        track.coarse_tiles_samples = hits
        return hits 

    def pixel_hit_finding(self, track, pixel_timeseries, nonoise = False, **kwargs):
        """
        This method of hit finding records the charge on a pixel
        to a capacitor buffer at each time tick.
        If the total charge collected is above a threshold,
        digitize each measurement
        """
        hits = []
        
        for pixel_key, timeseries in pixel_timeseries.items():
            pixel_center = (pixel_key[0], pixel_key[1])
            cell_trigger_t = pixel_key[2]

            time_ticks, interval_charge = timeseries
            
            discrim_charge = torch.sum(interval_charge)+torch.poisson(torch.tensor(self.readout_config['pixels']['noise']).float())
            threshold = self.readout_config['pixels']['noise']*self.readout_config['pixels']['threshold_sigma']
            if discrim_charge > threshold:
                measured_charge = interval_charge + torch.poisson(self.readout_config['pixels']['noise']*torch.ones_like(interval_charge))
                
                for this_timestamp, this_measured_charge in zip(time_ticks, measured_charge):
                    this_z = this_timestamp*1.6e5
                    hits.append(PixelSample(pixel_center,
                                            this_timestamp.item(),
                                            this_z.item(),
                                            this_measured_charge.item()))
                                            # threshold_crossing_t.item(),
                                            # threshold_crossing_z.item(),
                                            # threshold_crossing_charge.item()))
        track.pixel_samples = hits
        return hits 

class LArPixModel (ReadoutModel):
    def tile_hit_finding(self, track, tile_timeseries, nonoise = False, **kwargs):
        """
        This method of hit finding simply looks for a quantity
        of charge above threshold within a given z-bin
        (corresponding to a clock_period*integration_length)
        """
        # TODO: fix logic for integration_length > 1
        hits = []
        
        for tile_center, timeseries in tile_timeseries.items():
            time_ticks, interval_charge = timeseries
            
            hold_length = self.readout_config['coarse_tiles']['integration_length']            
            
            # search along the bins until no more threshold crossings
            no_more_hits = False
            while not no_more_hits:
                # window_charge = np.convolve(interval_charge,
                #                             np.ones(hold_length))
                window_charge = torch.conv_tbc(interval_charge[:,None,None],
                                               torch.ones(hold_length,1,1),
                                               bias = torch.zeros(1),
                                               pad = hold_length-1)[:,0,0]
                window_charge = window_charge[hold_length-1:]
                
                threshold = self.readout_config['coarse_tiles']['noise']*self.readout_config['coarse_tiles']['threshold_sigma']

                threshold_crossing_mask = window_charge > threshold
                threshold_crossing_mask *= interval_charge > 0

                if torch.any(threshold_crossing_mask):
                    hit_index = threshold_crossing_mask.nonzero()[0][0]
                    
                    threshold_crossing_t = time_ticks[hit_index]
                    threshold_crossing_z = threshold_crossing_t*1.6e5 # is there a better way to do this?
                    # TODO: also, get that from physics params
                    
                    threshold_crossing_charge = window_charge[hit_index]
                    if not nonoise:
                        threshold_crossing_charge += torch.poisson(torch.tensor(self.readout_config['coarse_tiles']['noise']).float())

                    interval_charge[:hit_index+hold_length] = 0

                    hits.append(CoarseGridSample(tile_center,
                                                 threshold_crossing_t.item(),
                                                 threshold_crossing_z.item(),
                                                 threshold_crossing_charge.item()))
                else:
                    no_more_hits = True

        track.coarse_tiles_samples = hits
        return hits 

    def pixel_hit_finding(self, track, pixel_timeseries, nonoise = False, **kwargs):
        """
        This method of hit finding simply looks for a quantity
        of charge above threshold within a given z-bin
        (corresponding to a clock_period*integration_length)
        """
        hits = []
        
        for pixel_key, timeseries in pixel_timeseries.items():
            pixel_center = (pixel_key[0], pixel_key[1])
            cell_trigger_t = pixel_key[2]

            time_ticks, interval_charge = timeseries

            hold_length = self.readout_config['pixels']['integration_length']            

            # search along the bins until no more threshold crossings
            no_hits = False
            while not no_hits:
                window_charge = torch.conv_tbc(interval_charge[:,None,None],
                                               torch.ones(hold_length,1,1),
                                               bias = torch.zeros(1),
                                               pad = hold_length-1)[:,0,0]
                window_charge = window_charge[hold_length-1:]

                threshold = self.readout_config['pixels']['noise']*self.readout_config['pixels']['threshold_sigma']
                threshold_crossing_mask = window_charge > threshold
                threshold_crossing_mask *= interval_charge > 0
                
                if torch.any(threshold_crossing_mask):
                    hit_index = threshold_crossing_mask.nonzero()[0][0]

                    threshold_crossing_t = time_ticks[hit_index]
                    threshold_crossing_z = threshold_crossing_t*1.6e5 # is there a better way to do this?
                    threshold_crossing_charge = window_charge[hit_index]

                    if not nonoise:
                        # add quiescent noise
                        # threshold_crossing_charge += np.random.normal(scale = self.readout_config['pixels']['noise'])
                        threshold_crossing_charge += torch.poisson(torch.tensor(self.readout_config['pixels']['noise']).float())

                    interval_charge[:hit_index+hold_length] = 0

                    hits.append(PixelSample(pixel_center,
                                            threshold_crossing_t.item(),
                                            threshold_crossing_z.item(),
                                            threshold_crossing_charge.item()))
                else:
                    no_hits = True

        track.pixel_samples = hits
        return hits 

class DetectorModel:
    def __init__(self,
                 detector_params = default_detector_params,
                 physics_params = default_physics_params,
                 readout_params = default_readout_params):
        self.detector_params = detector_params
        self.physics_params = physics_params
        self.readout_params = readout_params
        self.readout_model = GAMPixModel(readout_params)
        # self.readout_model = LArPixModel(readout_params)
 
    def simulate(self, track, **kwargs):
        self.drift(track, **kwargs)
        self.readout(track, **kwargs)
        
    def drift(self, sampled_track, **kwargs):
        # drift the charge samples from their input position
        # to the anode position as defined by detector_params
        # save the drifted positions to te track

        # TODO: a more complete way to describe the anode geometry
        # i.e., specify a plane and assume drift direction is shortest
        # path to that plane
        # anode_z = self.detector_params['anode']['z']
        anode_z = self.readout_params['anode']['z_range'][0]

        input_position = sampled_track.raw_track['4vec'][:,:3]
        input_charges = sampled_track.raw_track['charge']

        # position is disturbed by diffusion
        drift_distance = input_position[:,2] - anode_z

        # mask all points which are behind the anode
        region_mask = drift_distance > 0
        region_position = input_position[region_mask]
        region_charges = input_charges[region_mask]

        drift_distance = drift_distance[region_mask]

        # TODO: implement better drift model (maybe a functional response model)
        drift_time = drift_distance/self.physics_params['charge_drift']['drift_speed'] # s
        # D accordint to https://lar.bnl.gov/properties/trans.html#diffusion-l
        # sigma = sqrt(2*D*t)

        # use the nominal drift time to calculate diffusion
        # then, add the appropriate arrival time dispersion later
        sigma_transverse = torch.sqrt(2*self.physics_params['charge_drift']['diffusion_transverse']*drift_time)
        sigma_longitudinal = torch.sqrt(2*self.physics_params['charge_drift']['diffusion_longitudinal']*drift_time)
                                        
        diffusion_sigma = torch.stack((sigma_transverse,
                                       sigma_transverse,
                                       sigma_longitudinal,
                                       )).T
        
        drifted_positions = torch.normal(region_position,
                                         diffusion_sigma*np.ones_like(region_position))

        # charge is diminished by attenuation
        drifted_charges = region_charges*np.exp(-drift_time/self.physics_params['charge_drift']['electron_lifetime'])

        # add dispersion to the arrival of charge due to longitudinal diffusion
        time_dispersion = (drifted_positions[:, 2] - region_position[:, 2])/self.physics_params['charge_drift']['drift_speed'] 
        
        arrival_times = drift_time + sampled_track.raw_track['4vec'][region_mask,3] + time_dispersion
            
        # might also include a sub-sampling step?
        # in case initial sampling is not fine enough

        sampled_track.drifted_track = {'position': drifted_positions,
                                       'charge': drifted_charges,
                                       'times': arrival_times}
        return 

    def readout(self, drifted_track, **kwargs):
        # apply the readout simulation to the drifted track
        # this is defined by the GAMPixModel object

        self.readout_model.electronics_simulation(drifted_track, **kwargs)

        return



from gampixpy.tracks import Track
from gampixpy.config import default_physics_params
from gampixpy.units import *

import numpy as np
import torch
import particle

meta_dtype =  np.dtype([("event id", "u4"),
                        ("primary energy", "f4"),
                        ("deposited charge", "f4"),
                        ("vertex x", "f4"),
                        ("vertex y", "f4"),
                        ("vertex z", "f4"),
                        ("theta", "f4"),
                        ("phi", "f4"),
                        ("primary length", "f4"),
                        ],
                       align = True)

class InputParser:
    def __init__(self, input_filename, sequential_sampling = True, physics_config = default_physics_params):
        self.physics_config = physics_config
        
        self.input_filename = input_filename

        self.sampling_order = []

        self.open_file_handle()
        self.generate_sample_order(sequential_sampling)
        
    def __iter__(self):
        for sample_index in self.sampling_order:
            yield sample_index, self.get_sample(sample_index), self.get_meta(sample_index)

class SegmentParser (InputParser):
    def do_recombination(self, dE, dx, dEdx, mode = 'box', **kwargs):
        E_field = self.physics_config['charge_drift']['drift_field']
        LAr_density = self.physics_config['material']['density']

        if mode == 'box':
            box_beta = self.physics_config['box_model']['box_alpha']
            box_alpha = self.physics_config['box_model']['box_beta']

            csi = box_beta*dEdx/(E_field*LAr_density)
            recomb = torch.max(torch.stack([torch.zeros_like(dE),
                                            torch.log(box_alpha + csi)/csi]),
                               dim = 0)[0]

        elif mode == 'birks':
            birks_ab = self.physics_config['birks_model']['birks_ab']
            birks_kb = self.physics_config['birks_model']['birks_kb']

            recomb = birks_ab/(1 + birks_kb*dEdx/(E_field * LAr_density))

        else:
            raise ValueError("Invalid recombination mode: must be 'physics.BOX' or 'physics.BRIKS'")

        if torch.any(torch.isnan(recomb)):
            raise RuntimeError("Invalid recombination value")

        w_ion = self.physics_config['material']['w']
        charge_yield_per_energy = recomb/w_ion

        n_electrons = dE*charge_yield_per_energy
        return n_electrons
    
    def do_point_sampling(self, start_4vec, end_4vec,
                          dx, charge_per_segment,
                          sample_density = 1.e-1,
                          sample_normalization = 'charge',
                          **kwargs):
        # point sampling with a fixed number of samples per length
        # it may be faster to do sampling another way (test in future!)
        #  - sample with fixed amount of charge
        #  - sample with fixed number of samples per segment

        segment_interval = end_4vec - start_4vec

        if sample_normalization == 'charge':
            # here, sample_density is [samples/unit charge]
            samples_per_segment = (charge_per_segment*sample_density).int()
        elif sample_normalization == 'length':
            # here, sample_density is [samples/unit length]
            samples_per_segment = (dx*sample_density).int()

        sample_start = torch.repeat_interleave(start_4vec, samples_per_segment, dim = 0)
        sample_interval = torch.repeat_interleave(segment_interval, samples_per_segment, dim = 0)
        sample_parametric_distance = torch.cat(tuple(torch.linspace(0, 1, samples_per_segment[i])
                                                     for i in range(samples_per_segment.shape[0])))
        sample_4vec = sample_start + sample_interval*sample_parametric_distance[:,None]
        
        sample_charges = torch.repeat_interleave(charge_per_segment/samples_per_segment, samples_per_segment)

        return sample_4vec, sample_charges

class RooTrackerParser (SegmentParser):
    def open_file_handle(self, **kwargs):
        from ROOT import TFile, TG4Event

        self.file_handle = TFile(self.input_filename)
        self.inputTree = self.file_handle.Get("EDepSimEvents")

        self.event = TG4Event()
        self.inputTree.SetBranchAddress("Event", self.event)

    def generate_sample_order(self, sequential_sampling, **kwargs):
        n_images_per_file = self.inputTree.GetEntriesFast()
        if sequential_sampling:
            self.sampling_order = torch.arange(n_images_per_file)
        else:
            self.sampling_order = torch.randperm(n_images_per_file)

    def get_G4_sample(self, sample_index, **kwargs):
        self.inputTree.GetEntry(sample_index, **kwargs)

        start_4vec = torch.empty((0,4))
        end_4vec = torch.empty((0,4))
        dE = torch.empty((0))

        for container_name, hit_segments in self.event.SegmentDetectors:
            for segment in hit_segments:

                this_start_4vec = torch.tensor([segment.GetStart().X()*mm,
                                                segment.GetStart().Y()*mm,
                                                segment.GetStart().Z()*mm,
                                                segment.GetStart().T()*ns])
                this_end_4vec = torch.tensor([segment.GetStop().X()*mm,
                                              segment.GetStop().Y()*mm,
                                              segment.GetStop().Z()*mm,
                                              segment.GetStop().T()*ns])
                this_dE = torch.tensor([segment.GetEnergyDeposit()*MeV])
                
                start_4vec = torch.cat((this_start_4vec[None,:],
                                        start_4vec))
                end_4vec = torch.cat((this_end_4vec[None,:],
                                      end_4vec))
                dE = torch.cat((this_dE,
                                dE))

        displacement = start_4vec[:,:3] - end_4vec[:,:3]
        dx = torch.sum(displacement**2, dim = 1)
        dEdx = torch.where(dx > 0, dE/dx, 0)

        dQ = self.do_recombination(dE, dx, dEdx, **kwargs)
        charge_4vec, charge_values = self.do_point_sampling(start_4vec,
                                                            end_4vec,
                                                            dx, dQ,
                                                            **kwargs
                                                            )
        
        return Track(charge_4vec, charge_values)

    def get_G4_meta(self, sample_index, **kwargs):
        primary_vertex = self.event.Primaries[0] # assume only one primary for now

        vertex_x = primary_vertex.GetPosition().X()*mm
        vertex_y = primary_vertex.GetPosition().Y()*mm
        vertex_z = primary_vertex.GetPosition().Z()*mm
        
        primary_trajectory = self.event.Trajectories[0]
        assert primary_trajectory.GetParentId() == -1

        pdg_code = primary_trajectory.GetPDGCode()
        mass = particle.Particle.from_pdgid(pdg_code).mass*MeV
        energy = primary_trajectory.GetInitialMomentum()[3]*MeV
        momentum = [primary_trajectory.GetInitialMomentum()[i]*MeV for i in range(3)]
        kinetic_energy = energy - mass

        theta = np.arctan2(momentum[1], momentum[0])
        phi = np.arctan2(np.sqrt(momentum[0]**2 + momentum[1]**2), momentum[2])
        
        meta_array = np.array([(sample_index,
                                kinetic_energy,
                                -1, # charge undefined
                                vertex_x, vertex_y, vertex_z,
                                theta, phi,
                                -1, # primary length undefined
                                )],
                              dtype = meta_dtype)
        return meta_array

    def get_sample(self, index, **kwargs):
        return self.get_G4_sample(index, **kwargs)

    def get_meta(self, index, **kwargs):
        return self.get_G4_meta(index, **kwargs)

class EdepSimParser (SegmentParser):
    # Unit conventions for edepsim inputs:
    # distance: cm
    # energy: MeV
    def open_file_handle(self, **kwargs):
        import h5py
        self.file_handle = h5py.File(self.input_filename, **kwargs)

    def generate_sample_order(self, sequential_sampling, **kwargs):
        unique_event_ids = np.unique(self.file_handle['trajectories']['eventID']).astype(np.int32)
        # unique_event_ids = torch.tensor(unique_event_ids, dtype = torch.int32)
        n_images_per_file = len(unique_event_ids)
        if sequential_sampling:
            self.sampling_order = torch.tensor(unique_event_ids)
        else:
            self.sampling_order = torch.tensor(unique_event_ids[torch.randperm(n_images_per_file)])
        
    def get_edepsim_event(self, sample_index, **kwargs):
        segment_mask = self.file_handle['segments']['eventID'] == sample_index.item()
        event_segments = self.file_handle['segments'][segment_mask]

        trajectory_mask = self.file_handle['trajectories']['eventID'] == sample_index.item()
        event_trajectories = self.file_handle['trajectories'][trajectory_mask]

        start_4vec = torch.tensor((event_segments['x_start']*cm,
                                   event_segments['y_start']*cm,
                                   event_segments['z_start']*cm,
                                   event_segments['t_start']*ns,
                                   )).T
        end_4vec = torch.tensor((event_segments['x_end']*cm,
                                 event_segments['y_end']*cm,
                                 event_segments['z_end']*cm,
                                 event_segments['t_end']*ns,
                                 )).T
        dE = torch.tensor(event_segments['dE']*MeV)

        displacement = start_4vec[:,:3] - end_4vec[:,:3]
        dx = torch.sum(displacement**2, dim = 1)
        dEdx = torch.where(dx > 0, dE/dx, 0)

        dQ = self.do_recombination(dE, dx, dEdx, **kwargs)
        charge_4vec, charge_values = self.do_point_sampling(start_4vec,
                                                            end_4vec,
                                                            dx, dQ,
                                                            **kwargs
                                                            )
        return Track(charge_4vec, charge_values)
    
    def get_edepsim_meta(self, sample_index, **kwargs):
        trajectory_mask = self.file_handle['trajectories']['eventID'] == sample_index.item()
        event_trajectories = self.file_handle['trajectories'][trajectory_mask]
        primary_trajectory = event_trajectories[event_trajectories['parentID'] == -1]

        pdg_code = primary_trajectory['pdgId']
        mass = particle.Particle.from_pdgid(pdg_code).mass # MeV/c^2
        momentum = primary_trajectory['pxyz_start'] # MeV/c
        kinetic_energy = np.sqrt(np.power(mass, 2) + np.sum(np.power(momentum, 2))) - mass

        vertex = primary_trajectory['xyz_start'][0]
        init_momentum = primary_trajectory['pxyz_start'][0]

        theta = np.arctan2(init_momentum[1], init_momentum[0])
        phi = np.arctan2(np.sqrt(init_momentum[0]**2 + init_momentum[1]**2), init_momentum[2])
        
        meta_array = np.array([(sample_index,
                                kinetic_energy,
                                -1, # charge undefined
                                vertex[0], vertex[1], vertex[2],
                                theta, phi,
                                -1, # primary length undefined
                                )],
                              dtype = meta_dtype)
        return meta_array
        
    def get_sample(self, index, **kwargs):
        return self.get_edepsim_event(index, **kwargs)

    def get_meta(self, index, **kwargs):
        return self.get_edepsim_meta(index, **kwargs)

class MarleyParser (SegmentParser):
    # BROKEN
    # revisit this later
    def open_file_handle(self):
        from ROOT import TFile

        self.file_handle = TFile(self.input_filename)
        self.inputTree = self.file_handle.Get("edep")

    def generate_sample_order(self, sequential_sampling):
        event_ids = []
        for entry in self.inputTree:
            event_ids.append(entry.event)
        unique_event_ids = np.unique(event_ids)
        n_images_per_file = self.inputTree.GetEntriesFast()
        if sequential_sampling:
            self.sampling_order = unique_event_ids
        else:
            self.sampling_order = np.random.choice(unique_event_ids,
                                                   len(unique_event_ids),
                                                   replace = False)
    
    def get_G4_sample(self, sample_index):
        segment_dtype = np.dtype([("x_start", "f4"),
                                  ("y_start", "f4"),
                                  ("z_start", "f4"),
                                  ("x_end", "f4"),
                                  ("y_end", "f4"),
                                  ("z_end", "f4"),
                                  ("dE", "f4"),
                                  ("dx", "f4"),
                                  ("dEdx", "f4")],
                                align = True)
        segment_array = np.array([], dtype = segment_dtype)

        for entry in self.inputTree:
            if entry.event == sample_index:
                x_start = entry.startX
                y_start = entry.startY
                z_start = entry.startZ
                x_end = entry.endX
                y_end = entry.endY
                z_end = entry.endZ

                x_d = x_end - x_start
                y_d = y_end - y_start
                z_d = z_end - z_start
                dx = np.sqrt(x_d**2 + y_d**2 + z_d**2)
                
                dE = entry.dE
                dEdx = dE/dx if dx > 0 else 0

                this_segment_array = np.array([(x_start,
                                                y_start,
                                                z_start,
                                                x_end,
                                                y_end,
                                                z_end,
                                                dE,
                                                dx,
                                                dEdx)],
                                              dtype = segment_dtype)
                segment_array = np.concatenate((this_segment_array,
                                                segment_array))

        charge_per_segment = self.do_recombination(segment_array)
        charge_points, charge_values = self.do_point_sampling(segment_array, charge_per_segment)

        return Track(charge_points, charge_values)

    def get_G4_meta(self, sample_index):
        meta_array = np.array([(sample_index,
                                -1, # KE undefined
                                -1, # charge undefined
                                -1, -1, -1, # vertex point undefined
                                -1, -1, # primary attitude undefined
                                -1, # primary length undefined
                                )],
                              dtype = meta_dtype)
        return meta_array

    def get_sample(self, index):
        return self.get_G4_sample(index)

    def get_meta(self, index):
        return self.get_G4_meta(index)

class MarleyCSVParser (SegmentParser):
    def open_file_handle(self):
        self.col_names = ('run', 'subrun', 'event',
                          'isSignal', 'pdgCode', 'trackID',
                          'motherID', 'startE', 'dE',
                          'startX', 'startY',
                          'startZ', 'startT',
                          'endX', 'endY',
                          'endZ', 'endT',
                          )
        self.col_types = ('i4', 'i4', 'i4',
                          'b1', 'i4', 'i4',
                          'i4', 'f4', 'f4',
                          'f4', 'f4',
                          'f4', 'f4',
                          'f4', 'f4',
                          'f4', 'f4',
                          )

        self.data_table = np.loadtxt(self.input_filename,
                                     skiprows = 1, # first row is a header column
                                     delimiter = ',', 
                                     dtype = {'names': self.col_names,
                                              'formats': self.col_types},
                                     )

    def generate_sample_order(self, sequential_sampling):

        unique_event_ids = np.unique(self.data_table['event']).astype(np.int32)
        if sequential_sampling:
            self.sampling_order = torch.tensor(unique_event_ids)
        else:
            self.sampling_order = torch.tensor(unique_event_ids[torch.randperm(len(unique_event_ids))])

    def get_CSV_sample(self, sample_index):
        event_mask = self.data_table['event'] == sample_index
        event_rows = self.data_table[event_mask]
        
        segment_dtype = np.dtype([("x_start", "f4"),
                                  ("y_start", "f4"),
                                  ("z_start", "f4"),
                                  ("t_start", "f4"),
                                  ("x_end", "f4"),
                                  ("y_end", "f4"),
                                  ("z_end", "f4"),
                                  ("t_end", "f4"),
                                  ("dE", "f4"),
                                  ("dx", "f4"),
                                  ("dEdx", "f4")],
                                 align = True)

        start_4vec = torch.tensor((event_rows['startY']*mm,
                                   event_rows['startZ']*mm,
                                   event_rows['startX']*mm,
                                   event_rows['startT']*ns)).T
        end_4vec = torch.tensor((event_rows['endY']*mm,
                                 event_rows['endZ']*mm,
                                 event_rows['endX']*mm,
                                 event_rows['endT']*ns)).T
        
        dE = torch.tensor(event_rows['dE']*MeV)

        displacement = start_4vec[:,:3] - end_4vec[:,:3]
        dx = torch.sum(displacement**2, dim = 1)
        dEdx = torch.where(dx > 0, dE/dx, 0)

        dQ = self.do_recombination(dE, dx, dEdx)
        charge_4vec, charge_values = self.do_point_sampling(start_4vec,
                                                            end_4vec,
                                                            dx, dQ,
                                                            )
        return Track(charge_4vec, charge_values)

    def get_CSV_meta(self, sample_index):
        event_mask = self.data_table['event'] == sample_index
        event_rows = self.data_table[event_mask]

        kinetic_energy = -1
        charge = -1
        vertex = [-1, -1, -1]
        theta = -1
        phi = -1
        
        meta_array = np.array([(sample_index,
                                kinetic_energy,
                                charge, # charge undefined
                                vertex[0], vertex[1], vertex[2],
                                theta, phi,
                                -1, # primary length undefined
                                )],
                              dtype = meta_dtype)
        return meta_array

    def get_sample(self, index):
        return self.get_CSV_sample(index)

    def get_meta(self, index):
        return self.get_CSV_meta(index)

class PenelopeParser (InputParser):
    # BROKEN
    # revisit this later
    
    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    
    def open_file_handle(self):
        # import h5py
        # self.file_handle = h5py.File(self.input_filename)
        return 

    def generate_sample_order(self, sequential_sampling):
        # n_images_per_file = len(np.unique(self.file_handle['trajectories']['eventID']))
        return 
        
    def get_penelope_sample(self):
        print ("wowzers")
        print (self.input_filename)
        # do the magic that lets you read from a penelope file
        return None

    def get_sample(self, index):
        return self.get_penelope_sample(index)
    
    def get_meta(self, index):
        return None

parser_dict = {'root': RooTrackerParser,
               'edepsim': EdepSimParser,
               'marley': MarleyParser,
               }



import numpy as np
import torch

from gampixpy import tracks
from gampixpy.input_parsing import meta_dtype

class Generator:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __iter__(self):
        while True:
            yield self.get_sample(), self.get_meta()
    
class PointSource (Generator):
    def generate_sample_params(self):
        self.n_samples_per_point = 100000

        self.x_init = self.kwargs['x_range'][0] + (self.kwargs['x_range'][1] - self.kwargs['x_range'][0])*np.random.random()
        self.y_init = self.kwargs['y_range'][0] + (self.kwargs['y_range'][1] - self.kwargs['y_range'][0])*np.random.random()
        self.z_init = self.kwargs['z_range'][0] + (self.kwargs['z_range'][1] - self.kwargs['z_range'][0])*np.random.random()
        self.t_init = self.kwargs['t_range'][0] + (self.kwargs['t_range'][1] - self.kwargs['t_range'][0])*np.random.random()
        self.q_init = self.kwargs['q_range'][0] + (self.kwargs['q_range'][1] - self.kwargs['q_range'][0])*np.random.random()

    def get_sample(self):
        self.generate_sample_params()

        charge_4vec = torch.stack(self.n_samples_per_point*[[self.x_init,
                                                             self.y_init,
                                                             self.z_init,
                                                             self.t_init,
                                                             ]])
        charge_values = torch.tensor(self.n_samples_per_point*[self.q_init/self.n_samples_per_point,
                                                               ])                                 
        
        return tracks.Track(charge_4vec, charge_values)

    def get_meta(self):
        meta_array = np.array([(0, 0,
                                self.q_init,
                                self.x_init,
                                self.y_init,
                                self.z_init,
                                0,
                                0,
                                0)],
                              dtype = meta_dtype)
                
        return meta_array 

class LineSource (Generator):
    def generate_sample_params(self):
        self.n_samples_per_segment = 100000

        self.x_init = self.kwargs['x_range'][0] + (self.kwargs['x_range'][1] - self.kwargs['x_range'][0])*np.random.random()
        self.y_init = self.kwargs['y_range'][0] + (self.kwargs['y_range'][1] - self.kwargs['y_range'][0])*np.random.random()
        self.z_init = self.kwargs['z_range'][0] + (self.kwargs['z_range'][1] - self.kwargs['z_range'][0])*np.random.random()
        self.q = self.kwargs['q_range'][0] + (self.kwargs['q_range'][1] - self.kwargs['q_range'][0])*np.random.random()

        # maybe direction parameters can be passed as kwargs?
        # for now, let's just always use a uniform distribution
        # over the unit sphere
        self.theta = 2*np.pi*np.random.random()
        self.phi = np.arccos(1 - 2*np.random.random())

        self.length = self.kwargs['length_range'][0] + (self.kwargs['length_range'][1] - self.kwargs['length_range'][0])*np.random.random()
        
    def get_sample(self):
        self.generate_sample_params()

        pos_vec = np.array([self.x_init,
                            self.y_init,
                            self.z_init])
        dir_vec = np.array([np.cos(self.theta)*np.sin(self.phi),
                            np.sin(self.theta)*np.sin(self.phi),
                            np.cos(self.phi),
                            ])
        
        charge_points = pos_vec + np.linspace(0, 1, self.n_samples_per_segment)[:,None]*dir_vec*self.length
        # uniform charge profile (subject to change?)
        charge_values = np.array(self.n_samples_per_segment*[self.q/self.n_samples_per_segment,
                                                           ])                                 
        
        return tracks.Track(charge_points, charge_values)

    def get_meta(self):
        meta_array = np.array([(0, 0,
                                self.q,
                                self.x_init,
                                self.y_init,
                                self.z_init,
                                self.theta,
                                self.phi,
                                self.length)],
                              dtype = meta_dtype)
                
        return meta_array 



import math
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

class EventDisplay:
    MAX_POINTS_PLOTTED = 1e4
    def __init__(self, track):
        self.track_object = track
        self.init_fig()
        
    def init_fig(self):
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111, projection = '3d')

    def show(self):
        plt.show()

    def save(self, outfile):
        self.fig.savefig(outfile)

    def plot_raw_track(self, **kwargs):

        n_points = self.track_object.raw_track['position'].shape[0]
        if n_points > self.MAX_POINTS_PLOTTED:
            reduction_factor = math.ceil(n_points/self.MAX_POINTS_PLOTTED)
            self.ax.scatter(self.track_object.raw_track['position'][::reduction_factor,0],
                            self.track_object.raw_track['position'][::reduction_factor,1],
                            self.track_object.raw_track['position'][::reduction_factor,2],
                            c = np.log(self.track_object.raw_track['charge'][::reduction_factor]),
                            **kwargs,
                            )
        else:
            self.ax.scatter(self.track_object.raw_track['position'][:,0],
                            self.track_object.raw_track['position'][:,1],
                            self.track_object.raw_track['position'][:,2],
                            c = np.log(self.track_object.raw_track['charge'][:]),
                            **kwargs,
                            )

        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'z (drift) [cm]')
            
    def plot_drifted_track(self, **kwargs):

        n_points = self.track_object.drifted_track['position'].shape[0]
        if n_points > self.MAX_POINTS_PLOTTED:
            reduction_factor = math.ceil(n_points/self.MAX_POINTS_PLOTTED)
            self.ax.scatter(self.track_object.drifted_track['position'][::reduction_factor,0],
                            self.track_object.drifted_track['position'][::reduction_factor,1],
                            self.track_object.drifted_track['position'][::reduction_factor,2],
                            c = np.log(self.track_object.drifted_track['charge'][::reduction_factor]),
                            **kwargs,
                            )
        else:
            self.ax.scatter(self.track_object.drifted_track['position'][:,0],
                            self.track_object.drifted_track['position'][:,1],
                            self.track_object.drifted_track['position'][:,2],
                            c = np.log(self.track_object.drifted_track['charge'][:]),
                            **kwargs,
                            )

        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'z (drift) [cm]')

    def plot_drifted_track_timeline(self, **kwargs):

        n_points = self.track_object.drifted_track['position'].shape[0]
        if n_points > self.MAX_POINTS_PLOTTED:
            reduction_factor = math.ceil(n_points/self.MAX_POINTS_PLOTTED)
            self.ax.scatter(self.track_object.drifted_track['position'][::reduction_factor,0],
                            self.track_object.drifted_track['position'][::reduction_factor,1],
                            self.track_object.drifted_track['times'][::reduction_factor],
                            c = np.log(self.track_object.drifted_track['charge'][::reduction_factor]),
                            **kwargs,
                            )
        else:
            self.ax.scatter(self.track_object.drifted_track['position'][:,0],
                            self.track_object.drifted_track['position'][:,1],
                            self.track_object.drifted_track['times'][:],
                            c = np.log(self.track_object.drifted_track['charge'][:]),
                            **kwargs,
                            )

        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'z (drift) [cm]')

    def plot_coarse_tile_measurement(self, readout_config):
        # TODO: implement!

        for this_hit in self.track_object.coarse_tiles_samples:
            cell_center_xy = this_hit.coarse_cell_pos
            # cell_center_z = this_hit.coarse_measurement_time
            cell_center_t = this_hit.coarse_measurement_time
            v = 1.6e5
            cell_center_z = cell_center_t*v
            cell_measurement = this_hit.coarse_cell_measurement

            x_bounds = [cell_center_xy[0] - 0.5*readout_config['coarse_tiles']['pitch'],
                        cell_center_xy[0] + 0.5*readout_config['coarse_tiles']['pitch']]
            y_bounds = [cell_center_xy[1] - 0.5*readout_config['coarse_tiles']['pitch'],
                        cell_center_xy[1] + 0.5*readout_config['coarse_tiles']['pitch']]
            z_bounds = [cell_center_z,
                        cell_center_z + v*readout_config['coarse_tiles']['clock_interval']*readout_config['coarse_tiles']['integration_length']]
            
            
            bottom_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0]],
                                    [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                    [z_bounds[0], z_bounds[0], z_bounds[0], z_bounds[0], z_bounds[0]]]).T
            top_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0]],
                                 [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                 [z_bounds[1], z_bounds[1], z_bounds[1], z_bounds[1], z_bounds[1]]]).T
            left_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[0], x_bounds[0], x_bounds[0]],
                                  [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                  [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            right_face = np.array([[x_bounds[1], x_bounds[1], x_bounds[1], x_bounds[1], x_bounds[1]],
                                   [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                   [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            back_face = np.array([[x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0], x_bounds[0]],
                                  [y_bounds[0], y_bounds[0], y_bounds[0], y_bounds[0], y_bounds[0]],
                                  [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            front_face = np.array([[x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0], x_bounds[0]],
                                   [y_bounds[1], y_bounds[1], y_bounds[1], y_bounds[1], y_bounds[1]],
                                   [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            faces = [bottom_face,
                     top_face,
                     left_face,
                     right_face,
                     back_face,
                     front_face,
                     ]
            # self.ax.scatter(this_hit.coarse_cell_id[0],
            #                 this_hit.coarse_cell_id[1],
            #                 this_hit.coarse_measurement_time,
            #                 c = this_hit.coarse_cell_measurement,
            #                 )
            self.ax.add_collection3d(Poly3DCollection(faces, facecolors = 'cyan', linewidths=1, edgecolors = 'k', alpha = 0.25))
            
        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'z (drift) [cm]')

    def plot_coarse_tile_measurement_timeline(self, readout_config):
        # TODO: implement!

        xlim = []
        ylim = []
        zlim = []
        
        for this_hit in self.track_object.coarse_tiles_samples:
            cell_center_xy = this_hit.coarse_cell_pos
            # cell_center_z = this_hit.coarse_measurement_time
            cell_trigger_t = this_hit.coarse_measurement_time
            cell_measurement = this_hit.coarse_cell_measurement

            x_bounds = [cell_center_xy[0] - 0.5*readout_config['coarse_tiles']['pitch'],
                        cell_center_xy[0] + 0.5*readout_config['coarse_tiles']['pitch']]
            y_bounds = [cell_center_xy[1] - 0.5*readout_config['coarse_tiles']['pitch'],
                        cell_center_xy[1] + 0.5*readout_config['coarse_tiles']['pitch']]
            z_bounds = [cell_trigger_t,
                        cell_trigger_t + readout_config['coarse_tiles']['clock_interval']*readout_config['coarse_tiles']['integration_length']]
            
            bottom_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0]],
                                    [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                    [z_bounds[0], z_bounds[0], z_bounds[0], z_bounds[0], z_bounds[0]]]).T
            top_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0]],
                                 [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                 [z_bounds[1], z_bounds[1], z_bounds[1], z_bounds[1], z_bounds[1]]]).T
            left_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[0], x_bounds[0], x_bounds[0]],
                                  [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                  [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            right_face = np.array([[x_bounds[1], x_bounds[1], x_bounds[1], x_bounds[1], x_bounds[1]],
                                   [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                   [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            back_face = np.array([[x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0], x_bounds[0]],
                                  [y_bounds[0], y_bounds[0], y_bounds[0], y_bounds[0], y_bounds[0]],
                                  [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            front_face = np.array([[x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0], x_bounds[0]],
                                   [y_bounds[1], y_bounds[1], y_bounds[1], y_bounds[1], y_bounds[1]],
                                   [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            faces = [bottom_face,
                     top_face,
                     left_face,
                     right_face,
                     back_face,
                     front_face,
                     ]
            # self.ax.scatter(this_hit.coarse_cell_id[0],
            #                 this_hit.coarse_cell_id[1],
            #                 this_hit.coarse_measurement_time,
            #                 c = this_hit.coarse_cell_measurement,
            #                 )
            self.ax.add_collection3d(Poly3DCollection(faces, facecolors = 'cyan', linewidths=1, edgecolors = 'k', alpha = 0.25))
            
        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'arrival time [ns]')
        
    def plot_pixel_measurement(self, readout_config):
        
        for this_hit in self.track_object.pixel_samples:
            cell_center_xy = this_hit.pixel_pos
            cell_center_t = this_hit.hit_timestamp
            v = 1.6e5
            cell_center_z = cell_center_t*v
            cell_measurement = this_hit.hit_measurement

            x_bounds = [cell_center_xy[0] - 0.5*readout_config['pixels']['pitch'],
                        cell_center_xy[0] + 0.5*readout_config['pixels']['pitch']]
            y_bounds = [cell_center_xy[1] - 0.5*readout_config['pixels']['pitch'],
                        cell_center_xy[1] + 0.5*readout_config['pixels']['pitch']]
            z_bounds = [cell_center_z,
                        cell_center_z + v*readout_config['pixels']['clock_interval']*readout_config['pixels']['integration_length']]            
            
            bottom_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0]],
                                    [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                    [z_bounds[0], z_bounds[0], z_bounds[0], z_bounds[0], z_bounds[0]]]).T
            top_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0]],
                                 [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                 [z_bounds[1], z_bounds[1], z_bounds[1], z_bounds[1], z_bounds[1]]]).T
            left_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[0], x_bounds[0], x_bounds[0]],
                                  [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                  [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            right_face = np.array([[x_bounds[1], x_bounds[1], x_bounds[1], x_bounds[1], x_bounds[1]],
                                   [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                   [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            back_face = np.array([[x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0], x_bounds[0]],
                                  [y_bounds[0], y_bounds[0], y_bounds[0], y_bounds[0], y_bounds[0]],
                                  [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            front_face = np.array([[x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0], x_bounds[0]],
                                   [y_bounds[1], y_bounds[1], y_bounds[1], y_bounds[1], y_bounds[1]],
                                   [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            faces = [bottom_face,
                     top_face,
                     left_face,
                     right_face,
                     back_face,
                     front_face,
                     ]
            self.ax.add_collection3d(Poly3DCollection(faces,
                                                      facecolors = 'orange',
                                                      # facecolors = [np.log10(this_hit.hit_measurement)/10, 0, 0],
                                                      linewidths=1,
                                                      edgecolors = 'k',
                                                      alpha = 0.25))
            
        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'z (drift) [cm]')

    def plot_pixel_measurement_timeline(self, readout_config):
        
        for this_hit in self.track_object.pixel_samples:
            cell_center_xy = this_hit.pixel_pos
            # cell_center_z = this_hit.hit_timestamp
            cell_trigger_t = this_hit.hit_timestamp
            cell_measurement = this_hit.hit_measurement
            
            x_bounds = [cell_center_xy[0] - 0.5*readout_config['pixels']['pitch'],
                        cell_center_xy[0] + 0.5*readout_config['pixels']['pitch']]
            y_bounds = [cell_center_xy[1] - 0.5*readout_config['pixels']['pitch'],
                        cell_center_xy[1] + 0.5*readout_config['pixels']['pitch']]
            z_bounds = [cell_trigger_t,
                        cell_trigger_t + readout_config['pixels']['clock_interval']*readout_config['pixels']['integration_length']]
            
            
            bottom_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0]],
                                    [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                    [z_bounds[0], z_bounds[0], z_bounds[0], z_bounds[0], z_bounds[0]]]).T
            top_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0]],
                                 [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                 [z_bounds[1], z_bounds[1], z_bounds[1], z_bounds[1], z_bounds[1]]]).T
            left_face = np.array([[x_bounds[0], x_bounds[0], x_bounds[0], x_bounds[0], x_bounds[0]],
                                  [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                  [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            right_face = np.array([[x_bounds[1], x_bounds[1], x_bounds[1], x_bounds[1], x_bounds[1]],
                                   [y_bounds[0], y_bounds[1], y_bounds[1], y_bounds[0], y_bounds[0]],
                                   [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            back_face = np.array([[x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0], x_bounds[0]],
                                  [y_bounds[0], y_bounds[0], y_bounds[0], y_bounds[0], y_bounds[0]],
                                  [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            front_face = np.array([[x_bounds[0], x_bounds[1], x_bounds[1], x_bounds[0], x_bounds[0]],
                                   [y_bounds[1], y_bounds[1], y_bounds[1], y_bounds[1], y_bounds[1]],
                                   [z_bounds[0], z_bounds[0], z_bounds[1], z_bounds[1], z_bounds[0]]]).T
            faces = [bottom_face,
                     top_face,
                     left_face,
                     right_face,
                     back_face,
                     front_face,
                     ]
            self.ax.add_collection3d(Poly3DCollection(faces,
                                                      facecolors = 'orange',
                                                      # facecolors = [np.log10(this_hit.hit_measurement)/10, 0, 0],
                                                      linewidths=1,
                                                      edgecolors = 'k',
                                                      alpha = 0.25))
            
        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'arrival time [ns]')
